"""
:mod:`zsl_jwt.auth.module`
--------------------------

This ZSL module reads the auth configuration and provides
:class:`zsl_jwt.auth.configuration.AuthConfiguration` and an instance of
:class:`zsl_jwt.auth.service.AuthenticationService`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import *  # NOQA

from injector import Module
from injector import provides
from zsl import Config
from zsl import inject
from zsl.utils.import_helper import fetch_class

from zsl_jwt.auth.configuration import AuthConfiguration
from zsl_jwt.auth.service import AuthenticationService


class AuthModule(Module):
    AUTH_CONFIG_NAME = 'AUTH'

    @provides(AuthConfiguration)
    @inject(config=Config)
    def provide_jwt_configuration(self, config):
        # type: (Config) -> AuthConfiguration
        """
        Returns the AuthConfiguration.

        :param config: Injected. Configuration object
        :return: Current authentication/authorization configuration.
        """
        return config[AuthModule.AUTH_CONFIG_NAME]

    @provides(AuthenticationService)
    @inject(auth_config=AuthConfiguration)
    def provide_authentication_service(self, auth_config):
        # type: (AuthConfiguration)->AuthenticationService
        """
        Returns the used authentication service.

        :param auth_config: Injected. Auth configuration.
        :return: The authentication service.
        """

        if isinstance(auth_config.authentication_service_class, str):
            service_class = fetch_class(auth_config.authentication_service_class)
        else:
            service_class = auth_config.authentication_service_class

        @inject(service=service_class)
        def fetch_authentication_service(service):
            return service

        return fetch_authentication_service()



"""
:mod:`zsl_jwt.configuration`
-------------------------

The configuration of the authentication.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import *  # NOQA


class AuthConfiguration(object):
    """
    Auth module configuration. It holds the string identifying the
    authentication service.
    """

    def __init__(self, authentication_service_class):
        # type: (str)->None
        self._authentication_service_class = authentication_service_class

    @property
    def authentication_service_class(self):
        # type:()->str
        """
        The string identifying the class implementing the
        :class:`zsl_jwt.auth.service.AuthenticationService` which is used
        for login controller.
        """
        return self._authentication_service_class



"""
:mod:`zsl_jwt.auth.service`
---------------------------

The abstraction of authentication service and user information.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from abc import ABCMeta
from abc import abstractmethod
from builtins import *  # NOQA
from typing import Any  # NOQA
from typing import Callable  # NOQA
from typing import Generic  # NOQA
from typing import List  # NOQA
from typing import Set  # NOQA
from typing import Tuple  # NOQA
from typing import TypeVar  # NOQA

from zsl.db.model.app_model import AppModel

from zsl_jwt.codec import decode
from zsl_jwt.configuration import DEFAULT_PROFILE_NAME


class AuthenticationService(object):
    """
    The service used for verifying username and password and
    querying the user information.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def verify_password(self, username, password):
        # type:(str, str)->bool
        """Verifies if the username, password combination is valid.
        Returns true iff it is, False otherwise. It should not raise
        exceptions."""
        pass

    @abstractmethod
    def get_user_information(self, username):
        # type:(str)->Tuple[List[str], Any]
        """
        Returns the user information for the given username.
        :param username:
        :return: Tuple with the first element being the list of roles
                 (list of strings) and a user information, may be ``None``.
        """
        pass


T = TypeVar('T')


class StandardUserInformation(AppModel, Generic[T]):
    """Standard user information - contains username,
    roles (list of strings) and a user object, if wanted (may be ``None``)"""

    def __init__(self, username, roles, user_object):
        # type:(str, List[str], T)->None
        self._username = username
        self._roles = set(roles)
        self._user_object = user_object

        super(StandardUserInformation, self).__init__({})

    def get_attributes(self):
        return {
            'roles': list(self._roles),
            'username': self.username,
            'user_object': self._user_object
        }

    @property
    def username(self):
        # type: ()->str
        return self._username

    @property
    def roles(self):
        # type: ()->Set[str]
        return self._roles

    def is_in_role(self, role):
        # type:(str)->bool
        return role in self._roles

    @property
    def user_object(self):
        # type:()->T
        return self._user_object


def create_standard_user_information(username, roles, user_object):
    # type: (str, List[str], Any)->StandardUserInformation
    """Creates the user information/representation from the given parameters."""
    return StandardUserInformation(username, roles, user_object)


def decode_to_standard_user_information(token, user_object_class=None, profile=DEFAULT_PROFILE_NAME):
    # type: (str, Callable)->StandardUserInformation
    """Creates the user information/representation from the given auth token."""
    payload = decode(token, profile)
    if user_object_class is not None:
        payload['user_object'] = user_object_class(payload['user_object'])
    return StandardUserInformation(**payload)



import json
from unittest.case import TestCase

from zsl.application.containers.container import IoCContainer
from zsl.application.modules.context_module import DefaultContextModule
from zsl.application.modules.logger_module import LoggerModule
from zsl.db.model.app_model import AppModel
from zsl.task.job_context import WebJobContext
from zsl.testing.zsl import ZslTestCase
from zsl.testing.zsl import ZslTestConfiguration

from zsl_jwt.auth.configuration import AuthConfiguration
from zsl_jwt.auth.controller import ERROR_CODE_INVALID_CREDENTIALS
from zsl_jwt.auth.controller import authenticate
from zsl_jwt.auth.module import AuthModule
from zsl_jwt.auth.service import AuthenticationService
from zsl_jwt.auth.service import decode_to_standard_user_information
from zsl_jwt.configuration import JWTConfiguration
from zsl_jwt.module import JWTModule


class MyAppModel(AppModel):
    pass


class TestAuthenticationService(AuthenticationService):
    def get_user_information(self, username):
        return [['role'], MyAppModel({'id': 5})]

    def verify_password(self, username, password):
        return username == 'john' and password == 'doe'


class ZslAuthContainer(IoCContainer):
    logger = LoggerModule
    context = DefaultContextModule
    jwt_module = JWTModule
    auth_module = AuthModule()


class TestAuth(ZslTestCase, TestCase):
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name='zsl_jwt_encode_test',
        container=ZslAuthContainer,
        config_object={
            JWTModule.JWT_CONFIG_NAME: JWTConfiguration('secret'),
            AuthModule.AUTH_CONFIG_NAME: AuthConfiguration(TestAuthenticationService)
        }
    )

    INVALID_CREDENTIALS_RESPONSE = {"code": "INVALID_CREDENTIALS",
                                    "message": "Username and password can not be verified."}

    def testAuthenticate(self):
        WebJobContext('', '', '', None, None)
        self.assertEqual(ERROR_CODE_INVALID_CREDENTIALS,
                         authenticate('john', 'john').code,
                         "Credentials john:john must be invalid")
        self.assertTrue(isinstance(authenticate('john', 'doe'), dict),
                        'Credentials john:doe should be valid.')
        self.assertTrue('token' in authenticate('john', 'doe'),
                        'Credentials john:doe should be valid.')

    def testDecodeAuthToken(self):
        token = authenticate('john', 'doe')['token']
        user_info = decode_to_standard_user_information(token, MyAppModel)
        self.assertEqual('john', user_info.username)
        self.assertEqual({'role'}, user_info.roles)
        self.assertIsInstance(user_info.user_object, MyAppModel)
        self.assertEqual(5, user_info.user_object.id)



"""
:mod:`zsl_jwt.controller`
-------------------------

Contains the login function.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import http
from builtins import *  # NOQA
from typing import Any  # NOQA

from zsl import inject
from zsl.application.error_handler import ErrorResponse
from zsl.task.job_context import StatusCodeResponder
from zsl.task.job_context import add_responder
from zsl.utils.http import get_http_status_code_value

from zsl_jwt.auth.service import AuthenticationService
from zsl_jwt.auth.service import StandardUserInformation  # NOQA
from zsl_jwt.auth.service import create_standard_user_information
from zsl_jwt.decorators import jwt_output

ERROR_CODE_INVALID_CREDENTIALS = 'INVALID_CREDENTIALS'
ERROR_MESSAGE_INVALID_CREDENTIALS = "Username and password can not be verified."


@inject(authentication_service=AuthenticationService)
def authenticate(username, password, authentication_service):
    # type: (str, str, AuthenticationService)->Any
    """
    The "login" function, from the given username/password returns the JWT.
    Use only as a delegate function in handling of tasks.

    :param username:
    :param password:
    :param authentication_service: Injected. Uses
           :class:`zsl_jwt.auth.service.AuthenticationService` to query
           the database.
    :return: error response if the credentials are invalid or jwt.
    """
    # type:(str, str, AuthenticationService)->Any
    if not authentication_service.verify_password(username, password):
        return _respond_with_error()

    return _respond_with_jwt(username, authentication_service)


def _respond_with_error():
    # type: ()->ErrorResponse
    add_responder(StatusCodeResponder(get_http_status_code_value(http.client.FORBIDDEN)))
    return ErrorResponse(ERROR_CODE_INVALID_CREDENTIALS, ERROR_MESSAGE_INVALID_CREDENTIALS)


@jwt_output()
def _respond_with_jwt(username, authentication_service):
    # type:(str, AuthenticationService)->StandardUserInformation
    roles, user_object = authentication_service.get_user_information(username)
    return create_standard_user_information(username, roles, user_object).get_attributes()



"""
:mod:`zsl_jwt.codec`
--------------------

The module provides the two main functions :func:`.encode` and :func:`.decode`
which encode and decode the given payload.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
from builtins import *  # NOQA
from typing import Any  # NOQA
from typing import Dict  # NOQA

import jwt
from zsl import Injected
from zsl import inject
from zsl.db.model.app_model_json_encoder import AppModelJSONEncoder
from zsl.errors import ZslError

from zsl_jwt.configuration import DEFAULT_PROFILE_NAME
from zsl_jwt.configuration import JWTConfiguration

#: The list of JWT claim names.
CLAIMS = ('exp', 'nbf', 'iss', 'aud', 'iat')


class ZslJwtError(ZslError):
    """Main exception type raised from this module."""
    pass


class ZslJwtInvalidAudienceError(ZslJwtError):
    """When the audience of the token does not match the
    audience of the profile used to decode the token."""
    pass


class ZslJwtExpiredSignatureError(ZslJwtError):
    pass


class ZslJwtInvalidNbfClaimError(ZslJwtError):
    pass


@inject(jwt_configuration=JWTConfiguration)
def encode(payload, profile=DEFAULT_PROFILE_NAME, jwt_configuration=Injected):
    # type: (Dict[str, str], str, JWTConfiguration, str)->str
    """
    Encodes the payload.
    :param payload: The payload to be encoded.
    :param jwt_configuration: The JWT configuration, it is injected.
    :param algorithm: Algorithm name. A reasonable default is provided.
    :param profile: The JWT profile to be used. See
                    :class:`zsl_jwt.configuration.JWTProfile`.
    :return: The JWT token.
    """
    payload = _append_claims(payload, profile)
    profile = jwt_configuration[profile]
    token = jwt.encode(payload, profile.secret, json_encoder=AppModelJSONEncoder,
                       algorithm=profile.algorithm)
    token = token.decode('ascii')
    return token


@inject(jwt_configuration=JWTConfiguration)
def decode(token, profile=DEFAULT_PROFILE_NAME, jwt_configuration=Injected):
    # type: (str, str, JWTConfiguration)->Dict[str, Any]
    """
    Decodes the encoded token.

    :param token: The encoded token.
    :param jwt_configuration: The configuration, injected.
    :param profile: The profile name used for encoding.
    :return: The decoded payload.
    """
    try:
        profile = jwt_configuration[profile]
        payload = jwt.decode(token, profile.secret, audience=profile.audience)
    except jwt.InvalidAudienceError as e:
        raise ZslJwtInvalidAudienceError(e)
    except jwt.ExpiredSignatureError as e:
        raise ZslJwtExpiredSignatureError(e)
    except jwt.ImmatureSignatureError as e:
        raise ZslJwtInvalidNbfClaimError(e)
    except Exception as e:
        raise ZslJwtError(e)
    payload = _remove_claims(payload)
    return payload


@inject(jwt_configuration=JWTConfiguration)
def _append_claims(payload, profile, jwt_configuration):
    # type: (Dict[str, str], str, JWTConfiguration)->Dict[str, str]
    for claim in CLAIMS:
        if claim in payload:
            raise ZslJwtError("Claim '%s' found in payload. "
                              "Please use payloads without "
                              "claims - %s.".format(claim, CLAIMS))

    payload = payload.copy()
    profile = jwt_configuration[profile]
    payload['exp'] = datetime.datetime.utcnow() + profile.expiration
    payload['nbf'] = datetime.datetime.utcnow() + profile.not_before
    payload['iss'] = profile.issuer
    payload['aud'] = profile.audience
    payload['iat'] = datetime.datetime.utcnow()
    return payload


def _remove_claims(payload):
    for claim in CLAIMS:
        del payload[claim]
    return payload



"""
:mod:`zsl_jwt.decorators`
-------------------------

The module provides a decorator which turns the function's output into a JWT.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import *  # NOQA
from functools import wraps
from typing import Callable  # NOQA

from zsl import Config
from zsl import inject

from zsl_jwt.codec import encode
from zsl_jwt.configuration import DEFAULT_PROFILE_NAME


def jwt_output(profile=DEFAULT_PROFILE_NAME, debug=None):
    # type:(str, str)->Callable
    """
    Uses the returned dictionary as the JWT payload and returns
    a JSON with the token.

    :param profile: The JWT profile which is used.
    :param debug: The debug info - if true or the debug mode is on,
                  then the payload is in the returned JSON.
    :return: Dict with the 'token' key containing the created JWT.
    """

    def wrapper_function(f):
        @inject(config=Config)
        def get_config(config):
            # type: (Config)->Config
            return config

        @wraps(f)
        def wrapper(*args, **kwargs):
            payload = f(*args, **kwargs)
            d = {'token': encode(payload, profile=profile)}
            if debug or (debug is None and get_config().get('DEBUG', False)):
                d['payload'] = payload
            return d

        return wrapper

    return wrapper_function



"""
:mod:`zsl_jwt.configuration`
----------------------------
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
from builtins import *  # NOQA
from typing import Dict  # NOQA

from zsl.errors import ZslError

DEFAULT_PROFILE_NAME = 'default'
DEFAULT_ALGORITHM = 'HS256'

CLAIM_ISSUER = 'zsl_jwt'
CLAIM_AUDIENCE = 'zsl_jwt'
CLAIM_NOT_BEFORE = datetime.timedelta(seconds=0)
CLAIM_EXPIRATION = datetime.timedelta(hours=1)


class JWTConfiguration(object):
    """
    The main JWT configuration object. Consists of several token profiles
    used for encoding/decoding. See :class:`zsl_jwt.configuration.JWTProfile`.
    """

    def __init__(self, default_secret=None, default_profile=None,
                 profiles=None):
        # type:(str, JWTProfile, Dict[str, JWTProfile])->None
        if profiles is None:
            profiles = {}

        if DEFAULT_PROFILE_NAME not in profiles:
            if default_profile is None:
                if default_secret is None:
                    raise ZslError("JWT Secret or default profile not present.")
                default_profile = JWTProfile(default_secret, CLAIM_EXPIRATION,
                                             CLAIM_NOT_BEFORE, CLAIM_ISSUER,
                                             CLAIM_AUDIENCE)

            profiles[DEFAULT_PROFILE_NAME] = default_profile

        self._profiles = profiles

    def __getitem__(self, item):
        # type:(str)->JWTProfile
        return self._profiles[item]


class JWTProfile(object):
    """
    Each profile consits of its own
     - secret and all the JWT claims definitions:
     - audience: `str` - must match the audience in the decoding,
     - issuer name: `str` - just an information about the token issuer,
     - expiration: `datetime.timedelta` specifying when token becomes invalid,
     - not_before: `datetime.timedelta` specifying when token becomes valid,
     - algorithm: encryption algorithm used to create the token.
    """

    def __init__(self, secret, expiration=None, not_before=None, issuer=CLAIM_ISSUER,
                 audience=CLAIM_AUDIENCE, algorithm=DEFAULT_ALGORITHM):
        # type:(str, datetime.timedelta, datetime.timedelta, str, str, str)->None
        if expiration is None:
            expiration = CLAIM_EXPIRATION
        if not_before is None:
            not_before = CLAIM_NOT_BEFORE
        self._secret = secret
        self._expiration = expiration
        self._not_before = not_before
        self._issuer = issuer
        self._audience = audience
        self._algorithm = algorithm

    @property
    def algorithm(self):
        return self._algorithm

    @property
    def secret(self):
        """Key/secret used for encryption."""
        return self._secret

    @property
    def issuer(self):
        """The issuer name. This is a standard JWT claim."""
        return self._issuer

    @property
    def audience(self):
        """The audience of the token for which the token is intended.
        This must match the audience used for decoding. This is a standard
        JWT claim."""
        return self._audience

    @property
    def expiration(self):
        """The time interval specifying when the token becomes invalid.
        The token is valid until time now + expiration.
        This is a standard JWT claim."""
        return self._expiration

    @property
    def not_before(self):
        """The time interval specifying when the token becomes valid.
        The token is valid from time now + not_before.
        This is a standard JWT claim."""
        return self._not_before
